# -*- coding: utf-8 -*-
#
#    This file is part of PyNomo -
#    a program to create nomographs with Python (https://github.com/lefakkomies/pynomo)
#
#    Copyright (C) 2007-2019  Leif Roschier  <lefakkomies@users.sourceforge.net>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import math
import pyx
import numpy as np
import random
from copy import copy


class Axis_Wrapper:
    """
    class to wrap axis functionalities. Grid_wrapper and other classes
    will be derived from this.
    """

    def __init__(self, f, g, start, stop, sections=350):
        self.sections = sections  # how many sections are used for calculations
        self.f = f
        self.g = g
        self.start = start
        self.stop = stop
        # initial transformation coeffs
        self.set_transformation()
        self._calculate_points_()
        self.calc_length()
        self.calc_bound_box()

    def _calculate_points_(self):
        """
        calculates points and segments of the line
        """
        f = self.f
        g = self.g
        start = self.start
        stop = self.stop
        # du=math.fabs(start-stop)*1e-8
        du = math.fabs(start - stop) * 1e-12
        # approximate line length is found
        line_length_straigth = math.sqrt(
            (f(start) - f(stop)) ** 2 + (g(start) - g(stop)) ** 2)
        random.seed(0.1)  # so that mistakes always the same
        for dummy in range(100):
            first = random.uniform(start, stop)
            second = random.uniform(start, stop)
            temp = math.sqrt((f(first) - f(second)) ** 2 +
                             (g(first) - g(second)) ** 2)
            if temp > line_length_straigth:
                line_length_straigth = temp
                # print "length: %f"%line_length_straigth
        sections = self.sections  # about number of sections
        section_length = line_length_straigth / sections
        line = [(f(start), g(start))]
        line.append((f(start), g(start)))
        u = start
        count = 1
        while True:
            if u < stop:
                dx = (f(u + du) - f(u))
                dy = (g(u + du) - g(u))
                dl = math.sqrt(dx ** 2 + dy ** 2)
                if dl > 0:
                    delta_u = du * section_length / dl
                else:
                    delta_u = du  # dummy constant?
                # let's calculate actual length
                # and iterate until length is in factor 2 from target
                """
                while True:
                    delta_x=f(u+delta_u)-f(u)
                    delta_y=g(u+delta_u)-g(u)
                    delta_l=math.sqrt(delta_x**2+delta_y**2)

                    if delta_l>2.0*section_length:
                        delta_u=delta_u*0.999
                        #print "delta_u math.pienenee:%f"%delta_u
                    else:
                        if delta_l<section_length/2.0:
                            delta_u=delta_u*1.001
                            #print "delta_u kasvaa:%f"%delta_u
                    if delta_l<=2*section_length and delta_l>=0.5*section_length:
                        break
                """
                u += delta_u
                # print u,stop
                count = count + 1
                if u < stop:
                    line.append((f(u), g(u)))
            else:
                line.append((f(stop), g(stop)))
                # print count
                # print line
                break
        self.line = line
        # calculate sections
        sections = []
        for index, (x, y) in enumerate(line):
            if index > 1:
                sections.append((x, y, prev_x, prev_y))
            prev_x = x
            prev_y = y
        self.sections = sections

    def give_trafo_x(self, x, y):
        """
        transformed x-coordinate
        """
        return ((self.alpha1 * x + self.beta1 * y + self.gamma1)
                / (self.alpha3 * x + self.beta3 * y + self.gamma3))

    def give_trafo_y(self, x, y):
        """
        transformed y-coordinate
        """
        return ((self.alpha2 * x + self.beta2 * y + self.gamma2)
                / (self.alpha3 * x + self.beta3 * y + self.gamma3))

    def set_transformation(self, alpha1=1.0, beta1=0.0, gamma1=0.0,
                           alpha2=0.0, beta2=1.0, gamma2=0.0,
                           alpha3=0.0, beta3=0.0, gamma3=1.0):
        """
        sets the transformation for x,y points to be applied
        see funcs give_trafo_x and give_trafo_y for details
        """
        self.alpha1 = alpha1
        self.beta1 = beta1
        self.gamma1 = gamma1
        self.alpha2 = alpha2
        self.beta2 = beta2
        self.gamma2 = gamma2
        self.alpha3 = alpha3
        self.beta3 = beta3
        self.gamma3 = gamma3

    def calc_length(self):
        """
        calculates length of the basic line
        """
        length = 0
        for x1, y1, x2, y2 in self.sections:
            x1t = self.give_trafo_x(x1, y1)
            y1t = self.give_trafo_y(x1, y1)
            x2t = self.give_trafo_x(x2, y2)
            y2t = self.give_trafo_y(x2, y2)
            section = math.sqrt((x1t - x2t) ** 2 + (y1t - y2t) ** 2)
            length = length + section
        # print length
        self.length = length
        return length

    def plot_axis(self, c):
        """
        plots axis to pyx.canvas
        """
        x00, y00 = self.line[0]
        x0 = self.give_trafo_x(x00, y00)
        y0 = self.give_trafo_y(x00, y00)
        # print x0,y0
        line = pyx.path.path(pyx.path.moveto(x0, y0))
        for x, y in self.line:
            xt = self.give_trafo_x(x, y)
            yt = self.give_trafo_y(x, y)
            line.append(pyx.path.lineto(xt, yt))
        c.stroke(line, [pyx.style.linewidth.normal])

    def calc_bound_box(self):
        """
        calculates bounding box for axis
        """
        line = self.line
        (x_0, y_0) = line[0]
        x_left = self.give_trafo_x(x_0, y_0)
        y_top = self.give_trafo_y(x_0, y_0)
        x_right = self.give_trafo_x(x_0, y_0)
        y_bottom = self.give_trafo_y(x_0, y_0)
        for x, y in line:
            x_trafo = self.give_trafo_x(x, y)
            y_trafo = self.give_trafo_y(x, y)
            if x_trafo < x_left:
                x_left = x_trafo
            if x_trafo > x_right:
                x_right = x_trafo
            if y_trafo < y_bottom:
                y_bottom = y_trafo
            if y_trafo > y_top:
                y_top = y_trafo
        # print x_left,x_right,y_bottom,y_top
        # in case there is no area inside box, let's make
        # small in order to avoid math.singularities. These are
        # specifically for dual-axis stationary scales
        if x_left == x_right:
            x_left = x_right - 1e-2 * abs(y_top - y_bottom)
        if y_top == y_bottom:
            y_top = y_bottom + 1e-2 * abs(x_left - x_right)
        self.x_left = x_left
        self.x_right = x_right
        self.y_top = y_top
        self.y_bottom = y_bottom
        # print x_left,x_right,y_bottom,y_top
        return x_left, x_right, y_bottom, y_top

    def calc_highest_point(self):
        """
        calculates point with heighest y_value
        """
        line = self.line
        (x_0, y_0) = line[0]
        x_best = self.give_trafo_x(x_0, y_0)
        y_best = self.give_trafo_y(x_0, y_0)
        for x, y in line:
            x_trafo = self.give_trafo_x(x, y)
            y_trafo = self.give_trafo_y(x, y)
            if y_trafo > y_best:
                y_best = y_trafo
                x_best = x_trafo
        return x_best, y_best

    def calc_lowest_point(self):
        """
        calculates point with lowest y-value
        """
        line = self.line
        (x_0, y_0) = line[0]
        x_best = self.give_trafo_x(x_0, y_0)
        y_best = self.give_trafo_y(x_0, y_0)
        for x, y in line:
            x_trafo = self.give_trafo_x(x, y)
            y_trafo = self.give_trafo_y(x, y)
            if y_trafo < y_best:
                y_best = y_trafo
                x_best = x_trafo
        return x_best, y_best

    def calc_min_slope(self, x_ref, y_ref):
        """
        calculates minimum absolute slope of any point in axis and
        given point (x_ref,y_ref)
        """
        line = self.line
        (x_0, y_0) = line[0]
        x_best = self.give_trafo_x(x_0, y_0)
        y_best = self.give_trafo_y(x_0, y_0)
        slope_best = self._calc_slope_(x_best, y_best, x_ref, y_ref)
        for x, y in line:
            x_trafo = self.give_trafo_x(x, y)
            y_trafo = self.give_trafo_y(x, y)
            slope = self._calc_slope_(x_trafo, y_trafo, x_ref, y_ref)
            if slope < slope_best:
                y_best = y_trafo
                x_best = x_trafo
                slope_best = slope
        return x_best, y_best, slope_best

    def _calc_slope_(self, x1, y1, x2, y2):
        """
        calculates absolute value of slope between points (x1,y1) and (x2,y2)
        slope = dy/dx
        """
        dx = abs(x2 - x1)
        dy = abs(y1 - y2)
        if dx > 1e-9:  # close to zero
            return dy / dx
        else:
            return 1e120  # = big number


class Axes_Wrapper:
    """
    class to wrap axes group functionalities. For optimization of
    axes w.r.t. to paper
    """

    def __init__(self, paper_width=10.0, paper_height=10.0):
        self.paper_width = paper_width
        self.paper_height = paper_height
        self.paper_prop = paper_width / paper_height
        self.set_transformation()
        self.trafo_stack = []  # stack for transformation matrices
        self._add_transformation_()
        self.axes_list = []

    def add_axis(self, axis):
        """
        adds axis to the list
        """
        self.axes_list.append(axis)
        # if same instance of Axis_Wrapper is used in many Axes Wrapper
        # instances, let's reset
        self._set_transformation_to_all_axis_()

    def define_paper_size(self, paper_width=10.0, paper_height=10.0):
        """
        sets proportion of the paper
        """
        self.paper_width = paper_width
        self.paper_height = paper_height
        self.paper_prop = paper_width / paper_height

    def set_transformation(self, alpha1=1.0, beta1=0.0, gamma1=0.0,
                           alpha2=0.0, beta2=1.0, gamma2=0.0,
                           alpha3=0.0, beta3=0.0, gamma3=1.0):
        """
        sets the transformation for x,y points to be applied
        see funcs give_trafo_x and give_trafo_y for details
        """
        self.alpha1 = alpha1
        self.beta1 = beta1
        self.gamma1 = gamma1
        self.alpha2 = alpha2
        self.beta2 = beta2
        self.gamma2 = gamma2
        self.alpha3 = alpha3
        self.beta3 = beta3
        self.gamma3 = gamma3

    def _set_transformation_to_all_axis_(self):
        """
        sets current transformation to all axes
        """
        for axis in self.axes_list:
            axis.set_transformation(alpha1=self.alpha1, beta1=self.beta1, gamma1=self.gamma1,
                                    alpha2=self.alpha2, beta2=self.beta2, gamma2=self.gamma2,
                                    alpha3=self.alpha3, beta3=self.beta3, gamma3=self.gamma3)

    def _calc_bounding_box_(self):
        """
        calculates bounding box for the axes
        """
        x_left, x_right, y_bottom, y_top = self.axes_list[0].calc_bound_box()
        for axis in self.axes_list:
            x_left0, x_right0, y_bottom0, y_top0 = axis.calc_bound_box()
            if x_left0 < x_left:
                x_left = x_left0
            if x_right0 > x_right:
                x_right = x_right0
            if y_bottom0 < y_bottom:
                y_bottom = y_bottom0
            if y_top0 > y_top:
                y_top = y_top0
        self.x_left = x_left
        self.x_right = x_right
        self.y_top = y_top
        self.y_bottom = y_bottom
        self.Wt = self.x_right - self.x_left
        self.Ht = self.y_top - self.y_bottom
        return x_left, x_right, y_bottom, y_top

    def _calc_paper_area_(self):
        """
        calculates paper area needed to fit axes to
        given paper proportions
        """
        proportion = self.Wt / self.Ht
        if proportion >= self.paper_prop:
            W = self.Wt
            H = self.Wt / self.paper_prop
        else:  # proportion<self.paper_prop
            W = self.Ht * self.paper_prop
            H = self.Ht
        # self.multiplier_x=self.paper_width/self.Wt
        # self.multiplier_y=self.paper_height/self.Ht
        self.H = H
        self.W = W
        self.paper_area = W * H

    def _trafo_to_paper_(self):
        """
        transforms nomogram to paper proportions
        """
        self._calc_bounding_box_()
        """
        multiplier_x=self.paper_width/self.Wt
        multiplier_y=self.paper_height/self.Ht
        """
        x1 = self.x_left
        y1 = self.y_top
        x2 = self.x_left
        y2 = self.y_bottom
        x3 = self.x_right
        y3 = self.y_top
        x4 = self.x_right
        y4 = self.y_bottom
        x1d = 0.0
        y1d = self.paper_height
        x2d = 0.0
        y2d = 0.0
        x3d = self.paper_width
        y3d = self.paper_height
        x4d = self.paper_width
        y4d = 0.0

        alpha1, beta1, gamma1, alpha2, beta2, gamma2, alpha3, beta3, gamma3 = \
            self._calc_transformation_matrix_(x1, y1, x2, y2, x3, y3, x4, y4,
                                              x1d, y1d, x2d, y2d, x3d, y3d, x4d, y4d)
        """
        alpha1=multiplier_x
        beta1=0.0
        gamma1=0.0
        alpha2=0.0
        beta2=multiplier_y
        gamma2=0.0
        alpha3=0.0
        beta3=0.0
        gamma3=1.0
        """
        self._add_transformation_(alpha1=alpha1, beta1=beta1, gamma1=gamma1,
                                  alpha2=alpha2, beta2=beta2, gamma2=gamma2,
                                  alpha3=alpha3, beta3=beta3, gamma3=gamma3)
        self._set_transformation_to_all_axis_()

    def _calc_axes_length_sq_sum_(self):
        """
        calculates sum of axes squared lengths
        """
        length_sum_sq = 0.0
        for axis in self.axes_list:
            length_sum_sq = length_sum_sq + (axis.calc_length()) ** 2
        self.length_sum_sq = length_sum_sq
        return length_sum_sq

    def _calc_min_func_(self, params):
        """
        function to be minimized
        """
        # print params
        # print "."
        self._change_params_to_last_trafo_mat_(
            params)  # sets tranformation parameters
        self._set_transformation_to_all_axis_()  # applies trafo to every axis
        bb = self._calc_bounding_box_()
        self._calc_paper_area_()
        self._calc_axes_length_sq_sum_()
        opt_value = self.paper_area / self.length_sum_sq
        max_bb = max(bb)
        min_bb = min(bb)
        max_bb = max(max_bb, -min_bb)
        if max_bb > 1000.0:
            opt_value = max_bb
        return opt_value

    def optimize_transformation(self):
        """
        returns optimal transformation
        """
        x0 = [1.0, 0, 0, 0, 1.0, 0, 0, 0, 1.0]
        self._add_params_trafo_stack_(x0)
        print("starts optimizing...")
        np.fmin(self._calc_min_func_, x0, full_output=1, maxiter=2000)
        # self.alpha1=self.multiplier_x*self.alpha1
        # self.beta1=self.multiplier_x*self.beta1
        # self.gamma1=self.multiplier_x*self.gamma1
        # self.alpha2=self.multiplier_y*self.alpha2
        # self.beta2=self.multiplier_y*self.beta2
        # self.gamma2=self.multiplier_y*self.gamma2
        self._set_transformation_to_all_axis_()
        # self._calc_bounding_box_()
        # self._trafo_to_paper_()

    def fit_to_paper(self):
        """
        makes tranformation to fit to paper
        """
        self._trafo_to_paper_()

    def matrix_trafo(self, given_matrix):
        """
        adds explicite matrix transformation
        """
        gm = given_matrix
        self._add_transformation_(alpha1=gm[0][0], beta1=gm[0][1], gamma1=gm[0][2],
                                  alpha2=gm[1][0], beta2=gm[1][1], gamma2=gm[1][2],
                                  alpha3=gm[2][0], beta3=gm[2][1], gamma3=gm[2][2])
        self._set_transformation_to_all_axis_()

    def _plot_axes_(self, c):
        """
        prints axes for debugging purposes
        """
        for axis in self.axes_list:
            axis.plot_axis(c)

    def _print_result_pdf_(self, filename="axis_func_test.pdf"):
        """
        prints result pdf for debugging purposes
        """
        c = pyx.canvas.canvas()
        self._plot_axes_(c)
        c.writePDFfile(filename)
        # print original
        # print transformed

    def _calc_transformation_matrix_(self, x1, y1, x2, y2, x3, y3, x4, y4,
                                     x1d, y1d, x2d, y2d, x3d, y3d, x4d, y4d):
        """
        calculates transformation from orig_points (4 x-y pairs) to
        dest_points (4 x-y pairs):

            (x1,y1)     (x3,y3)          (x1d,y1d)      (x3d,y3d)
               |  polygon  |      ---->      |   polygon  |
            (x2,y2)     (x4,y4)          (x2d,y2d)      (x4d,y4d)
        """
        """
        o=orig_points
        x1,y1,x2,y2=o['x1'],o['y1'],o['x2'],o['y2']
        x3,y3,x4,y4=o['x3'],o['y3'],o['x4'],o['y4']
        d=dest_points
        x1d,y1d,x2d,y2d=d['x1'],d['y1'],d['x2'],d['y2']
        x3d,y3d,x4d,y4d=d['x3'],d['y3'],d['x4'],d['y4']
        """
        row1, const1 = self._make_row_(
            coordinate='x', coord_value=x2d, x=x2, y=y2)
        row2, const2 = self._make_row_(
            coordinate='y', coord_value=y2d, x=x2, y=y2)
        row3, const3 = self._make_row_(
            coordinate='x', coord_value=x1d, x=x1, y=y1)
        row4, const4 = self._make_row_(
            coordinate='y', coord_value=y1d, x=x1, y=y1)
        row5, const5 = self._make_row_(
            coordinate='x', coord_value=x4d, x=x4, y=y4)
        row6, const6 = self._make_row_(
            coordinate='y', coord_value=y4d, x=x4, y=y4)
        row7, const7 = self._make_row_(
            coordinate='x', coord_value=x3d, x=x3, y=y3)
        row8, const8 = self._make_row_(
            coordinate='y', coord_value=y3d, x=x3, y=y3)

        matrix = np.array([row1, row2, row3, row4, row5, row6, row7, row8])
        # print matrix
        b = np.array([const1, const2, const3, const4,
                      const5, const6, const7, const8])
        coeff_vector = np.linalg.solve(matrix, b)
        alpha1 = -1.0  # fixed
        beta1 = coeff_vector[0][0]
        gamma1 = coeff_vector[1][0]
        alpha2 = coeff_vector[2][0]
        beta2 = coeff_vector[3][0]
        gamma2 = coeff_vector[4][0]
        alpha3 = coeff_vector[5][0]
        beta3 = coeff_vector[6][0]
        gamma3 = coeff_vector[7][0]
        return alpha1, beta1, gamma1, alpha2, beta2, gamma2, alpha3, beta3, gamma3

    def give_trafo(self):
        return self.alpha1, self.beta1, self.gamma1, \
            self.alpha2, self.beta2, self.gamma2, \
            self.alpha3, self.beta3, self.gamma3

    def _make_row_(self, coordinate='x', x=1.0, y=1.0, coord_value=1.0):
        """ Utility to find transformation matrix. See eq.37,a
        in Allcock. We take \alpha_1=1. h=1.
        """
        # to make expressions shorter
        cv = coord_value
        if coordinate == 'x':
            row = np.array([y, 1, 0, 0, 0, -cv * x, -cv * y, -cv * 1])
            value = np.array([x])
        if coordinate == 'y':
            row = np.array([0, 0, x, y, 1, -cv * x, -cv * y, -cv * 1])
            value = np.array([0])
        return row, value

    def _find_polygon_horizontal_(self):
        """
        finds intersection of horizontal line "dropmath.ping" until it
        hit the highest point (1) of axes. Then line tilts to minimum angle
        by another point (2).
        Same for bottom points (3) and (4) with vice versa.
        """
        # let's find the point with largest y-value
        y_high = -1e120  # big negative number
        for idx, axis in enumerate(self.axes_list):
            x, y = axis.calc_highest_point()
            if y > y_high:
                y_high = y
                x_high = x
                high_idx = idx
        # let's find the point with smallest y-value
        y_low = 1e120  # big number
        for idx, axis in enumerate(self.axes_list):
            x, y = axis.calc_lowest_point()
            if y < y_low:
                y_low = y
                x_low = x
                low_idx = idx
        # let's find the top-line with minimum slope
        slope_high = 1e120  # big number
        for idx, axis in enumerate(self.axes_list):
            x, y, slope = axis.calc_min_slope(x_high, y_high)
            if slope < slope_high and not idx == high_idx:
                y_slope_high = y
                x_slope_high = x
                slope_high = slope
        # let's find the bottom-lne with minimum slope
        slope_low = 1e120  # big number
        for idx, axis in enumerate(self.axes_list):
            x, y, slope = axis.calc_min_slope(x_low, y_low)
            if slope < slope_low and not idx == low_idx:
                y_slope_low = y
                x_slope_low = x
                slope_low = slope
        """ let's set the points to a ractangle form (not self-intersecting)
            (x1,y1)   -  (x3,y3)
               |  polygon  |
            (x2,y2)   -  (x4,y4)
        """
        x1, y1, x2, y2 = x_high, y_high, x_low, y_low
        x3, y3, x4, y4 = x_slope_high, y_slope_high, x_slope_low, y_slope_low
        # print x_high, y_high, x_low, y_low, x_slope_high, y_slope_high, x_slope_low, y_slope_low
        if (x1 - x3) * (x2 - x4) < 0:
            x1, y1, x3, y3 = x3, y3, x1, y1
        if x3 < x1:
            x1, y1, x2, y2, x3, y3, x4, y4 = x3, y3, x4, y4, x1, y1, x2, y2
        return x1, y1, x2, y2, x3, y3, x4, y4

    def make_polygon_trafo(self):
        """
        transforms polygon according to:

        finds intersection of horizontal line "dropmath.ping" until it
        hit the highest point (1) of axes. Then line tilts to minimum angle
        by another point (2).
        Same for bottom points (3) and (4) with vice versa. This polygon will
        be transformed to the corners of the paper.

            (x1,y1)     (x3,y3)          (x1d,y1d)      (x3d,y3d)
               |  polygon  |      ---->      |   polygon  |
            (x2,y2)     (x4,y4)          (x2d,y2d)      (x4d,y4d)
        """
        # find the left polygon
        x1, y1, x2, y2, x3, y3, x4, y4 = self._find_polygon_horizontal_()
        # define right polygon
        x1d, y1d = x1, self.paper_height
        x2d, y2d = x2, 0.0
        x3d, y3d = x3, self.paper_height
        x4d, y4d = x4, 0.0

        c = pyx.canvas.canvas()
        self._plot_axes_(c)
        c.fill(pyx.path.circle(x1, y1, 0.02))
        c.text(x1 + 1, y1, '1')
        c.fill(pyx.path.circle(x2, y2, 0.03))
        c.text(x2 + 1, y2, '2')
        c.fill(pyx.path.circle(x3, y3, 0.04))
        c.text(x3 + 1, y3, '3')
        c.fill(pyx.path.circle(x4, y4, 0.05))
        c.text(x4 + 1, y4, '4')
        # c.writePDFfile('poly_debug.pdf')
        # print "polygon coords:"
        # print x1,y1,x2,y2,x3,y3,x4,y4
        # calculate transformation
        alpha1, beta1, gamma1, alpha2, beta2, gamma2, alpha3, beta3, gamma3 = \
            self._calc_transformation_matrix_(x1, y1, x2, y2, x3, y3, x4, y4,
                                              x1d, y1d, x2d, y2d, x3d, y3d, x4d, y4d)
        # apply transformation
        self._add_transformation_(alpha1, beta1, gamma1, alpha2, beta2, gamma2,
                                  alpha3, beta3, gamma3)
        self._set_transformation_to_all_axis_()  # update axis
        self._trafo_to_paper_()  # transforms to paper

    def _add_transformation_(self, alpha1=1.0, beta1=0.0, gamma1=0.0,
                             alpha2=0.0, beta2=1.0, gamma2=0.0,
                             alpha3=0.0, beta3=0.0, gamma3=1.0):
        """
        adds transformation to be applied as a basis.
        all transformation matrices are multiplied together
        """
        trafo_mat = np.array([[alpha1, beta1, gamma1],
                              [alpha2, beta2, gamma2],
                              [alpha3, beta3, gamma3]])
        self.trafo_stack.append(trafo_mat)
        self._calculate_total_trafo_mat_()  # update coeffs

    def _calculate_total_trafo_mat_(self):
        """
        calculates total transformation matrix and
        master coeffs self.alpha1,self.beta1,...
        """
        stack_copy = copy(self.trafo_stack)
        stack_copy.reverse()
        trafo_mat = stack_copy.pop()
        for matrix in stack_copy:
            trafo_mat = np.dot(trafo_mat, matrix)  # matrix multiplication
        self.alpha1 = trafo_mat[0][0]
        self.beta1 = trafo_mat[0][1]
        self.gamma1 = trafo_mat[0][2]
        self.alpha2 = trafo_mat[1][0]
        self.beta2 = trafo_mat[1][1]
        self.gamma2 = trafo_mat[1][2]
        self.alpha3 = trafo_mat[2][0]
        self.beta3 = trafo_mat[2][1]
        self.gamma3 = trafo_mat[2][2]

    def _change_params_to_last_trafo_mat_(self, params):
        """
        changes transformation coeffs from optimization params to
        last transformation matrix in stack
        """
        alpha1 = params[0]
        beta1 = params[1]
        gamma1 = params[2]
        alpha2 = params[3]
        beta2 = params[4]
        gamma2 = params[5]
        alpha3 = params[6]
        beta3 = params[7]
        gamma3 = params[8]
        self.trafo_stack.pop()
        self._add_transformation_(alpha1=alpha1, beta1=beta1, gamma1=gamma1,
                                  alpha2=alpha2, beta2=beta2, gamma2=gamma2,
                                  alpha3=alpha3, beta3=beta3, gamma3=gamma3)

    def _add_params_trafo_stack_(self, params):
        """
        appends transformation coeffs from (optimization) params to stack
        """
        alpha1 = params[0]
        beta1 = params[1]
        gamma1 = params[2]
        alpha2 = params[3]
        beta2 = params[4]
        gamma2 = params[5]
        alpha3 = params[6]
        beta3 = params[7]
        gamma3 = params[8]
        self._add_transformation_(alpha1=alpha1, beta1=beta1, gamma1=gamma1,
                                  alpha2=alpha2, beta2=beta2, gamma2=gamma2,
                                  alpha3=alpha3, beta3=beta3, gamma3=gamma3)

    def _calc_rotation_trafo_(self, angle):
        """
        returns rotation transformation. angle in degrees.
        """
        angle_r = angle / 180.0 * math.pi  # angle in radians
        alpha1 = math.cos(angle_r)
        beta1 = -math.sin(angle_r)
        gamma1 = 0.0
        alpha2 = math.sin(angle_r)
        beta2 = math.cos(angle_r)
        gamma2 = 0.0
        alpha3 = 0.0
        beta3 = 0.0
        gamma3 = 1.0
        return alpha1, beta1, gamma1, alpha2, beta2, gamma2, alpha3, beta3, gamma3

    def rotate_canvas(self, angle):
        """
        rotates pyx.canvas by angle degrees by adding a rotation matrix
        into trafo_stack
        """
        alpha1, beta1, gamma1, alpha2, \
            beta2, gamma2, alpha3, beta3, gamma3 = self._calc_rotation_trafo_(
                angle)
        self._add_transformation_(alpha1=alpha1, beta1=beta1, gamma1=gamma1,
                                  alpha2=alpha2, beta2=beta2, gamma2=gamma2,
                                  alpha3=alpha3, beta3=beta3, gamma3=gamma3)
        self._set_transformation_to_all_axis_()


if __name__ == '__main__':
    """
    testing
    """

    def f1(L):
        return (2 * (L * L - 8 * L - 5) / (3 * L * L + 2 * L + 7))

    def g1(L):
        return 10 * (8 * L * L + 12 * L - 8) / (3 * L * L + 2 * L + 7)

    def f2(L):
        return math.log10(L)

    def g2(L):
        return 10 ** (L)

    def f3(L):
        return -L

    def g3(L):
        return L - 1.0

    def f4(L):
        return 3.0

    def g4(L):
        return L

    def f5(L):
        return 6.0

    def g5(L):
        return L

    test1_ax = Axis_Wrapper(f1, g1, 0.5, 1.0)
    test2_ax = Axis_Wrapper(f2, g2, 0.5, 1.0)
    test3_ax = Axis_Wrapper(f3, g3, 0.0, 1.0)
    test4_ax = Axis_Wrapper(f4, g4, 0.0, 2.0)
    test5_ax = Axis_Wrapper(f5, g5, 0.0, 3.0)
    test3a_ax = Axis_Wrapper(f3, g3, 0.0, 1.0)
    test4a_ax = Axis_Wrapper(f4, g4, 0.0, 2.0)
    test5a_ax = Axis_Wrapper(f5, g5, 0.0, 3.0)
    test_wrap = Axes_Wrapper()
    # print test_wrap._calc_transformation_matrix_(1, 1, 1, 2, 1, 3, 1, 4, 3, 3, 3, 4, 3, 5, 3, 6)
    test_wrap.add_axis(test3_ax)
    test_wrap.add_axis(test4_ax)
    test_wrap.add_axis(test5_ax)
    # test_wrap._calc_min_func_([3.0,0,0,0,2,0,0,0,1])
    test_wrap._print_result_pdf_("original.pdf")
    # test_wrap.optimize_transformation()
    # test_wrap._print_result_pdf_("final.pdf")

    test_wrap1 = Axes_Wrapper()
    test_wrap1.add_axis(test3_ax)
    test_wrap1.add_axis(test4_ax)
    test_wrap1.add_axis(test5_ax)
    # test_wrap._calc_min_func_([3.0,0,0,0,2,0,0,0,1])
    test_wrap1.rotate_canvas(45.0)
    test_wrap1._print_result_pdf_("final_rot.pdf")

    test_wrap2 = Axes_Wrapper()
    test_wrap2.add_axis(test3a_ax)
    test_wrap2.add_axis(test4a_ax)
    test_wrap2.add_axis(test5a_ax)
    # test_wrap._calc_min_func_([3.0,0,0,0,2,0,0,0,1])
    test_wrap2.rotate_canvas(45.0)
    test_wrap2._print_result_pdf_("ini_poly.pdf")
    test_wrap2.make_polygon_trafo()
    test_wrap2._print_result_pdf_("final_poly.pdf")



# -*- coding: utf-8 -*-
#    This file is part of PyNomo -
#    a program to create nomographs with Python (https://github.com/lefakkomies/pynomo)
#
#    Copyright (C) 2007-2019  Leif Roschier  <lefakkomies@users.sourceforge.net>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .isopleth import Isopleth_Wrapper
from .nomo_wrapper import Nomo_Wrapper
from .nomo_wrapper import Nomo_Block_Type_1
from .nomo_wrapper import Nomo_Block_Type_2
from .nomo_wrapper import Nomo_Block_Type_3
from .nomo_wrapper import Nomo_Block_Type_4
from .nomo_wrapper import Nomo_Block_Type_5
from .nomo_wrapper import Nomo_Block_Type_6
from .nomo_wrapper import Nomo_Block_Type_7
from .nomo_wrapper import Nomo_Block_Type_8
from .nomo_wrapper import Nomo_Block_Type_9
from .nomo_wrapper import Nomo_Block_Type_10
from .nomo_axis import Nomo_Axis
from .nomo_axis import find_linear_ticks
from pprint import pprint

import pyx
import numpy as np


class Nomographer:
    """
    Top-level class to build nomographs
    """

    def __init__(self, params):
        """
        params hold all information to build the nomograph
        """
        self._check_params_(params)  # sets default values for misnp.sing keys
        wrapper = Nomo_Wrapper(params=params,
                               paper_width=params['paper_width'],
                               paper_height=params['paper_height'],
                               filename=params['filename'])
        blocks = []
        isopleths = Isopleth_Wrapper(params)
        for block_para in params['block_params']:
            # TYPE 1
            if block_para['block_type'] == 'type_1':
                self._check_block_type_1_params_(block_para)
                blocks.append(Nomo_Block_Type_1(mirror_x=block_para['mirror_x'],
                                                mirror_y=block_para['mirror_y']))
                self._check_axis_params_(block_para['f1_params'])
                self._check_axis_params_(block_para['f2_params'])
                self._check_axis_params_(block_para['f3_params'])
                blocks[-1].define_F1(block_para['f1_params'])
                blocks[-1].define_F2(block_para['f2_params'])
                blocks[-1].define_F3(block_para['f3_params'])
                blocks[-1].set_block(width=block_para['width'],
                                     height=block_para['height'],
                                     proportion=block_para['proportion'])
                wrapper.add_block(blocks[-1])
                isopleths.add_isopleth_block(blocks[-1], block_para)
            # TYPE 2
            if block_para['block_type'] == 'type_2':
                self._check_block_type_2_params_(block_para)
                blocks.append(Nomo_Block_Type_2(mirror_x=block_para['mirror_x'],
                                                mirror_y=block_para['mirror_y']))
                self._check_axis_params_(block_para['f1_params'])
                self._check_axis_params_(block_para['f2_params'])
                self._check_axis_params_(block_para['f3_params'])
                blocks[-1].define_F1(block_para['f1_params'])
                blocks[-1].define_F2(block_para['f2_params'])
                blocks[-1].define_F3(block_para['f3_params'])
                blocks[-1].set_block(width=block_para['width'],
                                     height=block_para['height'])
                wrapper.add_block(blocks[-1])
                isopleths.add_isopleth_block(blocks[-1], block_para)
            # TYPE 3
            if block_para['block_type'] == 'type_3':
                self._check_block_type_3_params_(block_para)
                blocks.append(Nomo_Block_Type_3(mirror_x=block_para['mirror_x'],
                                                mirror_y=block_para['mirror_y']))
                for axis_params in block_para['f_params']:
                    self._check_axis_params_(axis_params)
                    blocks[-1].add_F(axis_params)
                blocks[-1].set_block(width=block_para['width'],
                                     height=block_para['height'],
                                     reference_padding=block_para['reference_padding'],
                                     reference_titles=block_para['reference_titles'],
                                     reference_color=block_para['reference_color'])
                wrapper.add_block(blocks[-1])
                isopleths.add_isopleth_block(blocks[-1], block_para)
            # TYPE 4
            if block_para['block_type'] == 'type_4':
                self._check_block_type_4_params_(block_para)
                blocks.append(Nomo_Block_Type_4(mirror_x=block_para['mirror_x'],
                                                mirror_y=block_para['mirror_y']))
                self._check_axis_params_(block_para['f1_params'])
                self._check_axis_params_(block_para['f2_params'])
                self._check_axis_params_(block_para['f3_params'])
                self._check_axis_params_(block_para['f4_params'])
                blocks[-1].define_F1(block_para['f1_params'])
                blocks[-1].define_F2(block_para['f2_params'])
                blocks[-1].define_F3(block_para['f3_params'])
                blocks[-1].define_F4(block_para['f4_params'])
                blocks[-1].set_block(width=block_para['width'],
                                     height=block_para['height'],
                                     float_axis=block_para['float_axis'],
                                     padding=block_para['padding'],
                                     reference_color=block_para['reference_color'])
                wrapper.add_block(blocks[-1])
                isopleths.add_isopleth_block(blocks[-1], block_para)
            # TYPE 5
            if block_para['block_type'] == 'type_5':
                self._check_block_type_5_params_(block_para)
                blocks.append(Nomo_Block_Type_5(mirror_x=block_para['mirror_x'],
                                                mirror_y=block_para['mirror_y']))
                blocks[-1].define_block(block_para)
                blocks[-1].set_block()
                wrapper.add_block(blocks[-1])
                isopleths.add_isopleth_block(blocks[-1], block_para)
            # TYPE 6
            if block_para['block_type'] == 'type_6':
                self._check_block_type_6_params_(block_para)
                blocks.append(Nomo_Block_Type_6(mirror_x=block_para['mirror_x'],
                                                mirror_y=block_para['mirror_y']))
                blocks[-1].define(params1=block_para['f1_params'],
                                  params2=block_para['f2_params'])
                blocks[-1].set_block(width=block_para['width'],
                                     height=block_para['height'],
                                     type=block_para['type'],
                                     x_empty=block_para['x_empty'],
                                     y_empty=block_para['y_empty'],
                                     curve_const=block_para['curve_const'],
                                     ladder_color=block_para['ladder_color'])
                wrapper.add_block(blocks[-1])
                isopleths.add_isopleth_block(blocks[-1], block_para)
            # TYPE 7
            if block_para['block_type'] == 'type_7':
                self._check_block_type_7_params_(block_para)
                blocks.append(Nomo_Block_Type_7(mirror_x=block_para['mirror_x'],
                                                mirror_y=block_para['mirror_y']))
                self._check_axis_params_(block_para['f1_params'])
                self._check_axis_params_(block_para['f2_params'])
                self._check_axis_params_(block_para['f3_params'])
                blocks[-1].define_F1(block_para['f1_params'])
                blocks[-1].define_F2(block_para['f2_params'])
                blocks[-1].define_F3(block_para['f3_params'])
                blocks[-1].set_block(width_1=block_para['width_1'],
                                     angle_u=block_para['angle_u'],
                                     angle_v=block_para['angle_v'])
                wrapper.add_block(blocks[-1])
                isopleths.add_isopleth_block(blocks[-1], block_para)
            # TYPE 8
            if block_para['block_type'] == 'type_8':
                self._check_block_type_8_params_(block_para)
                blocks.append(Nomo_Block_Type_8(mirror_x=block_para['mirror_x'],
                                                mirror_y=block_para['mirror_y']))
                self._check_axis_params_(block_para['f_params'])
                blocks[-1].define_F(block_para['f_params'])
                blocks[-1].set_block(length=block_para['length'])
                wrapper.add_block(blocks[-1])
                isopleths.add_isopleth_block(blocks[-1], block_para)
            # TYPE 9
            if block_para['block_type'] == 'type_9':
                self._check_block_type_9_params_(block_para)
                blocks.append(Nomo_Block_Type_9(mirror_x=block_para['mirror_x'],
                                                mirror_y=block_para['mirror_y']))
                self._check_axis_params_(block_para['f1_params'])
                self._check_axis_params_(block_para['f2_params'])
                self._check_axis_params_(block_para['f3_params'])
                blocks[-1].define_determinant(block_para['f1_params'],
                                              block_para['f2_params'],
                                              block_para['f3_params'],
                                              transform_ini=block_para['transform_ini'])

                blocks[-1].set_block(width=block_para['width'],
                                     height=block_para['height'],
                                     ignore_transforms=block_para['ignore_transforms'])
                wrapper.add_block(blocks[-1])
                isopleths.add_isopleth_block(blocks[-1], block_para)
            # TYPE 10
            if block_para['block_type'] == 'type_10':
                self._check_block_type_10_params_(block_para)
                blocks.append(Nomo_Block_Type_10(mirror_x=block_para['mirror_x'],
                                                 mirror_y=block_para['mirror_y']))
                self._check_axis_params_(block_para['f1_params'])
                self._check_axis_params_(block_para['f2_params'])
                self._check_axis_params_(block_para['f3_params'])
                blocks[-1].define_F1(block_para['f1_params'])
                blocks[-1].define_F2(block_para['f2_params'])
                blocks[-1].define_F3(block_para['f3_params'])
                blocks[-1].set_block(width=block_para['width'],
                                     height=block_para['height'])
                wrapper.add_block(blocks[-1])
                isopleths.add_isopleth_block(blocks[-1], block_para)
            # always save a handle
            blocks[-1].ref_block_params = block_para
        wrapper.align_blocks()
        wrapper.build_axes_wrapper()  # build structure for transformations
        # do transformations
        for trafo in params['transformations']:
            if len(trafo) > 1:
                wrapper.do_transformation(method=trafo[0], params=trafo[1])
            else:
                wrapper.do_transformation(method=trafo[0])
        # transformations done
        c = pyx.canvas.canvas()
        if params['make_grid']:
            self._make_grid_(params, c)
        if params['pre_func'] is not None:
            params['pre_func'](c)
        if params['draw_lines']:
            self._draw_lines_(params, c)
        if params['draw_isopleths']:
            # draw isopleths
            isopleths.draw(c)
        else:  # calculate points
            for block in blocks:
                for atom in block.atom_stack:
                    # calculates lines (list of coordinates)
                    atom.calc_line_and_sections()
                    # pass
        # draw the nomogram
        wrapper.draw_nomogram(c, params['post_func'])
        self.blocks = blocks  # save for debugging
        for block in params['block_params']:
            if block['debug']:
                print("##### np.sinGLE BLOCK PARAMS #######")
                pprint(block)
        if params['debug']:
            print("##### MAIN PARAMS #######")
            pprint(params)
        self.wrapper = wrapper
        self.canvas = c

    def _make_grid_(self, params, c):
        """
        makes a grid to help position titles, etc.
        """
        axis_offset = 3.0  # how much scales are aside
        axis_color = pyx.color.cmyk.Brown
        Nomo_Axis(func_f=lambda u: u,
                  func_g=lambda u: 0.0 - axis_offset + u * 1e-5,
                  start=-axis_offset, stop=params['paper_width'] + axis_offset, turn=-1, title='',
                  tick_levels=3, tick_text_levels=2,
                  canvas=c, type='linear', side='right', axis_appear={'turn_relative': True,
                                                                      'axis_color': axis_color,
                                                                      'text_color': axis_color})
        Nomo_Axis(func_f=lambda u: u,
                  func_g=lambda u: params['paper_height'] +
                  axis_offset + u * 1e-5,
                  start=-axis_offset, stop=params['paper_width'] + axis_offset, turn=-1, title='',
                  tick_levels=3, tick_text_levels=2,
                  canvas=c, type='linear', side='left', axis_appear={'turn_relative': True,
                                                                     'axis_color': axis_color,
                                                                     'text_color': axis_color})
        Nomo_Axis(func_f=lambda u: 0.0 - axis_offset + u * 1e-5,
                  func_g=lambda u: u,
                  start=-axis_offset, stop=params['paper_height'] + axis_offset, turn=-1, title='',
                  tick_levels=3, tick_text_levels=2,
                  canvas=c, type='linear', side='left', axis_appear={'turn_relative': True,
                                                                     'axis_color': axis_color,
                                                                     'text_color': axis_color})
        Nomo_Axis(func_f=lambda u: params['paper_width'] + axis_offset + u * 1e-5,
                  func_g=lambda u: u,
                  start=-axis_offset, stop=params['paper_height'] + axis_offset, turn=-1, title='',
                  tick_levels=3, tick_text_levels=2,
                  canvas=c, type='linear', side='right', axis_appear={'turn_relative': True,
                                                                      'axis_color': axis_color,
                                                                      'text_color': axis_color})
        tick_0_list_v, tick_1_list_v, tick_2_list_v, tick_3_list_v, tick_4_list_v, \
            start_ax, stop_ax = find_linear_ticks(
                -axis_offset, params['paper_height'] + axis_offset)
        tick_0_list_h, tick_1_list_h, tick_2_list_h, tick_3_list_h, tick_4_list_h, \
            start_ax, stop_ax = find_linear_ticks(
                -axis_offset, params['paper_width'] + axis_offset)
        grid_color_0 = pyx.color.cmyk.Brown
        grid_color_1 = pyx.color.cmyk.Gray
        grid_color_2 = pyx.color.cmyk.Tan
        for tick in tick_0_list_v:
            c.stroke(pyx.path.line(-axis_offset, tick, params['paper_width'] + axis_offset, tick),
                     [grid_color_0, pyx.style.linewidth.THin])
        for tick in tick_1_list_v:
            c.stroke(pyx.path.line(-axis_offset, tick, params['paper_width'] + axis_offset, tick),
                     [grid_color_1, pyx.style.linewidth.THIN])
        for tick in tick_2_list_v:
            c.stroke(pyx.path.line(-axis_offset, tick, params['paper_width'] + axis_offset, tick),
                     [grid_color_2, pyx.style.linewidth.THIN])
        for tick in tick_0_list_h:
            c.stroke(pyx.path.line(tick, -axis_offset, tick, params['paper_height'] + axis_offset),
                     [grid_color_0, pyx.style.linewidth.THin])
        for tick in tick_1_list_h:
            c.stroke(pyx.path.line(tick, -axis_offset, tick, params['paper_height'] + axis_offset),
                     [grid_color_1, pyx.style.linewidth.THIN])
        for tick in tick_2_list_h:
            c.stroke(pyx.path.line(tick, -axis_offset, tick, params['paper_height'] + axis_offset),
                     [grid_color_2, pyx.style.linewidth.THIN])

    def _draw_lines_(self, params, c):
        """
        draws (isopleth) lines according to given coordinates
        """
        for line_defs in params['line_params']:
            # line style
            if 'line_style' in line_defs:
                line_style = line_defs['line_style']
            else:
                line_style = self.line_defs_default['line_style']
            # circle size
            if 'circle_size' in line_defs:
                circle_size = line_defs['circle_size']
            else:
                circle_size = self.line_defs_default['circle_size']
            # circle color
            if 'circle_color' in line_defs:
                circle_color = line_defs['circle_color']
            else:
                circle_color = self.line_defs_default['circle_color']
            # do lines and circles
            for line in line_defs['coords']:
                c.stroke(pyx.path.line(
                    line[0], line[1], line[2], line[3]), line_style)
                c.fill(pyx.path.circle(
                    line[0], line[1], circle_size), [circle_color])
                c.fill(pyx.path.circle(
                    line[2], line[3], circle_size), [circle_color])

    def _check_params_(self, params):
        """
        checks if main params ok and adds default values
        """
        self.line_defs_default = {'coords': [[0, 0, 1, 1], [2, 2, 3, 3]],
                                  'line_style': [pyx.color.cmyk.Black,
                                                 pyx.style.linewidth.thick,
                                                 pyx.style.linestyle.dashed],
                                  'circle_size': 0.0005,
                                  'circle_color': pyx.color.cmyk.Black,
                                  }
        params_default = {
            'filename': 'pynomo_default.pdf',
            'paper_height': 20.0,
            'paper_width': 20.0,
            # 'block_params':[test1_block1_params,test1_block2_params],
            'transformations': [('rotate', 0.01), ('scale paper',)],
            'title_color': pyx.color.rgb.black,
            'make_grid': False,
            'draw_lines': False,  # to draw manual lines
            'line_params': [self.line_defs_default],
            'pre_func': None,  # function(pyx.canvas) to draw first
            'post_func': None,  # function(pyx.canvas) to draw last
            'debug': False,
            'draw_isopleths': True,  # draws isopleths
            'isopleth_params': [{'color': 'Black',
                                 'linestyle': 'Dashed',
                                 'lineweight': 'thick',
                                 'circle_size': 0.05}],  # isopleth values
        }
        for key in params_default:
            if not key in params:
                params[key] = params_default[key]

    def _check_block_type_1_params_(self, params):
        """
        checks if block type 1 params ok and adds default values
        """
        params_default = {
            'mirror_x': False,
            'mirror_y': False,
            'width': 10.0,
            'height': 10.0,
            'proportion': 1.0,
            'debug': False,
            'isopleth_values': [],
        }
        for key in params_default:
            if not key in params:
                params[key] = params_default[key]

    def _check_block_type_2_params_(self, params):
        """
        checks if block type 2 params ok and adds default values
        """
        params_default = {
            'mirror_x': False,
            'mirror_y': False,
            'width': 10.0,
            'height': 10.0,
            'debug': False,
            'isopleth_values': [],
        }
        for key in params_default:
            if not key in params:
                params[key] = params_default[key]

    def _check_block_type_3_params_(self, params):
        """
        checks if block type 3 params ok and adds default values
        """
        params_default = {
            'mirror_x': False,
            'mirror_y': False,
            'width': 10.0,
            'height': 10.0,
            'reference_padding': 0.2,
            'reference_titles': [],
            'reference_color': pyx.color.rgb.black,
            'debug': False,
            'isopleth_values': [],
        }
        for key in params_default:
            if not key in params:
                params[key] = params_default[key]

    def _check_block_type_4_params_(self, params):
        """
        checks if block type 4 params ok and adds default values
        """
        params_default = {
            'mirror_x': False,
            'mirror_y': False,
            'width': 10.0,
            'height': 10.0,
            'float_axis': 'F1 or F2',
            'padding': 0.9,
            'reference_color': pyx.color.rgb.black,
            'debug': False,
            'isopleth_values': [],
        }
        for key in params_default:
            if not key in params:
                params[key] = params_default[key]

    def _check_block_type_5_params_(self, params):
        """
        checks if block type 5 params ok and adds default values
        """
        params_default = {
            'mirror_x': False,
            'mirror_y': False,
            'width': 10.0,
            'height': 10.0,
            # 'u_func':lambda u:u,
            # 'v_func':lambda x,v:x+v,
            # 'u_values':[10.0,15.0,20.0,25.0,30.0,40.0,50.0,60.0],
            # 'v_values':[1.0,2.0,3.0,4.0,5.0,6.0,7.0,8.0,9.0,10.0],
            'v_tick_side': 'left',
            'u_title': '',
            'v_title': '',
            'u_reference': False,  # manual labels
            'v_reference': False,
            'w_reference': False,
            'wd_reference': False,
            'wd_tick_levels': 0,
            'wd_tick_text_levels': 0,
            'wd_tag': 'none',
            'w_tick_levels': 0,
            'w_tick_text_levels': 0,
            'horizontal_guides': False,
            'debug': False,
            'isopleth_values': [],
        }
        for key in params_default:
            if not key in params:
                params[key] = params_default[key]

    def _check_block_type_6_params_(self, params):
        """
        checks if block type 6 params ok and adds default values
        """
        params_default = {
            'mirror_x': False,
            'mirror_y': False,
            'width': 10.0,
            'height': 10.0,
            'type': 'parallel',
            'x_empty': 0.2,
            'y_empty': 0.2,
            'curve_const': 0.0,
            'ladder_color': pyx.color.rgb.black,
            'debug': False,
            'isopleth_values': [],
        }
        for key in params_default:
            if not key in params:
                params[key] = params_default[key]

    def _check_block_type_7_params_(self, params):
        """
        checks if block type 7 params ok and adds default values
        """
        params_default = {
            'mirror_x': False,
            'mirror_y': False,
            'width_1': 10.0,
            'angle_u': 45.0,
            'angle_v': 45.0,
            'debug': False,
            'isopleth_values': [],
        }
        for key in params_default:
            if not key in params:
                params[key] = params_default[key]

    def _check_block_type_8_params_(self, params):
        """
        checks if block type 8 params ok and adds default values
        """
        params_default = {
            'mirror_x': False,
            'mirror_y': False,
            'length': 10.0,
            'debug': False,
            'isopleth_values': [],
        }
        for key in params_default:
            if not key in params:
                params[key] = params_default[key]

    def _check_block_type_9_params_(self, params):
        """
        checks if block type 9 params ok and adds default values
        """
        params_default = {
            'mirror_x': False,
            'mirror_y': False,
            'width': 10.0,
            'height': 10.0,
            'transform_ini': False,
            'grid': False,
            'v_texts_u_start': False,
            'v_texts_u_stop': True,
            'u_texts_v_start': False,
            'u_texts_v_stop': True,
            'debug': False,
            'isopleth_values': [],
            'ignore_transforms': False,
        }
        for key in params_default:
            if not key in params:
                params[key] = params_default[key]

    def _check_block_type_10_params_(self, params):
        """
        checks if block type 10 params ok and adds default values
        """
        params_default = {
            'mirror_x': False,
            'mirror_y': False,
            'width': 10.0,
            'height': 10.0,
            'debug': False,
            'isopleth_values': [],
        }
        for key in params_default:
            if not key in params:
                params[key] = params_default[key]

    def _check_axis_params_(self, params):
        """
        checks (TODO: if axis params ok) and adds default values
        """
        params_default = {
            'ID': 'none',  # to identify the axis
            'tag': 'none',  # for aligning block wrt others
            # 'u_min':0.1,
            # 'u_max':1.0,
            # 'F':lambda u:u, # x-coordinate
            # 'G':lambda u:u, # y-coordinate
            'title': '',
            'title_x_shift': 0.0,
            'title_y_shift': 0.25,
            'scale_type': 'linear',  # 'linear' 'log' 'manual point' 'manual line'
            'tick_levels': 4,
            'tick_text_levels': 3,
            'tick_side': 'right',
            'reference': False,
            'reference_padding': 0.20,  # fraction of reference line over other lines
            'manual_axis_data': {},
            'title_distance_center': 0.5,
            'title_opposite_tick': True,
            'title_draw_center': False,
            'grid': False
        }
        for key in params_default:
            if not key in params:
                params[key] = params_default[key]


if __name__ == '__main__':
    """
    tests
    """
    test1 = False
    if test1:
        test1_f1_para = {
            'u_min': 1.0,
            'u_max': 10.0,
            'function': lambda u: u,
            'title': 'F1',
            'tick_levels': 3,
            'tick_text_levels': 2,
        }
        test1_f2_para = {
            'u_min': 0.1,
            'u_max': 10.0,
            'function': lambda u: u,
            'title': 'F2',
            'tick_levels': 3,
            'tick_text_levels': 2,
            'scale_type': 'log',
        }
        test1_f3_para = {
            'u_min': 1.0,
            'u_max': 10.0,
            # 'function':lambda u:u*12.0,
            'function': lambda u: u,
            'title': 'F3',
            'tag': 'A',
            'tick_side': 'right'
        }
        test1_f4_para = {
            'u_min': -10.0,
            'u_max': 10.0,
            'function': lambda u: u,
            'title': 'F2',
            'tick_levels': 3,
            'tick_text_levels': 2,
        }
        test1_block1_params = {
            'block_type': 'type_1',
            'width': 10.0,
            'height': 10.0,
            'proportion': 0.5,
            'f1_params': test1_f1_para,
            'f2_params': test1_f2_para,
            'f3_params': test1_f3_para}

        test1_block7_params = {
            'block_type': 'type_7',
            'width_1': 10.0,
            'angle_u': 60.0,
            'angle_v': 60.0,
            'f1_params': test1_f1_para,
            'f2_params': test1_f2_para,
            'f3_params': test1_f3_para}

        test1_params = {
            'filename': 'test1.pdf',
            'paper_height': 20.0,
            'paper_width': 20.0,
            'block_params': [test1_block1_params, test1_block7_params],
            'transformations': [('rotate', 0.01), ('scale paper',)]
        }
        Nomographer(test1_params)

        test2_block2_params = {
            'block_type': 'type_2',
            'width': 10.0,
            'height': 10.0,
            'proportion': 0.5,
            'f1_params': test1_f1_para,
            'f2_params': test1_f2_para,
            'f3_params': test1_f3_para}

        test2_params = {
            'filename': 'test2.pdf',
            'paper_height': 20.0,
            'paper_width': 20.0,
            'block_params': [test2_block2_params],
            'transformations': [('rotate', 0.01), ('scale paper',)]
        }
        Nomographer(test2_params)

        test3_block3_params = {
            'block_type': 'type_3',
            'width': 10.0,
            'height': 10.0,
            'f_params': [test1_f1_para, test1_f1_para,
                         test1_f1_para, test1_f1_para, test1_f4_para]
        }

        test3_params = {
            'filename': 'test3.pdf',
            'paper_height': 20.0,
            'paper_width': 20.0,
            'block_params': [test3_block3_params],
            'transformations': [('rotate', 0.01), ('scale paper',)]
        }
        Nomographer(test3_params)

        test4_block4_params = {
            'block_type': 'type_4',
            'f1_params': test1_f1_para,
            'f2_params': test1_f2_para,
            'f3_params': test1_f3_para,
            'f4_params': test1_f1_para,
        }

        test4_params = {
            'filename': 'test4.pdf',
            'paper_height': 20.0,
            'paper_width': 20.0,
            'block_params': [test4_block4_params],
            'transformations': [('rotate', 0.01), ('scale paper',)]
        }

        Nomographer(test4_params)

        def f1(x, u):
            # return log(log(x/(x-u/100.0))/log(1+u/100.0))
            return np.log(np.log(x / (x - u / (100.0 * 12.0))) / np.log(1 + u / (100.0 * 12.0)))

        test5_block5_params = {
            'block_type': 'type_5',
            'u_func': lambda u: np.log(u * 12.0),
            'v_func': f1,
            'u_values': [10.0, 15.0, 20.0, 25.0, 30.0, 40.0, 50.0, 60.0],
            'v_values': [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0],
        }

        test5_params = {
            'filename': 'test5.pdf',
            'paper_height': 20.0,
            'paper_width': 20.0,
            'block_params': [test5_block5_params],
            'transformations': [('rotate', 0.01), ('scale paper',)]
        }

        Nomographer(test5_params)

        test6_block6_params = {
            'block_type': 'type_6',
            'f1_params': test1_f1_para,
            'f2_params': test1_f2_para,
        }
        test6_params = {
            'filename': 'test6.pdf',
            'paper_height': 20.0,
            'paper_width': 20.0,
            'block_params': [test6_block6_params],
            'transformations': [('rotate', 0.01), ('scale paper',)]
        }
        Nomographer(test6_params)

    F_start = -40.0
    F_stop = 90.0
    C_start = -40.0
    C_stop = 30.0

    def celcius(fahrenheit):
        return (fahrenheit - 32.0) / 1.8

    test8_f1_para = {
        'tag': 'A',
        'u_min': F_start,
        'u_max': F_stop,
        'function': lambda u: celcius(u),
        'title': r'$^\circ$ F',
        'tick_levels': 4,
        'tick_text_levels': 3,
        'align_func': celcius,
        'title_x_shift': 0.5
    }
    test8_f1c_para = {
        'tag': 'A',
        'u_min': F_start,
        'u_max': F_stop,
        'function': lambda u: celcius(u),
        'title': r'$^\circ$ F',
        'tick_levels': 4,
        'tick_text_levels': 3,
        'align_func': celcius,
        'title_x_shift': 0.5,
        'align_x_offset': 7.0
    }
    test8_f2_para = {
        'tag': 'A',
        'u_min': C_start,
        'u_max': C_stop,
        'function': lambda u: u,
        'title': r'$^\circ$ C',
        'tick_levels': 5,
        'tick_text_levels': 3,
        'scale_type': 'linear',
        'tick_side': 'left',
        'title_x_shift': -0.5
    }

    test8_block8a_params = {
        'block_type': 'type_8',
        'f_params': test8_f1_para
    }
    test8_block8c_params = {
        'block_type': 'type_8',
        'f_params': test8_f1c_para
    }
    test8_block8b_params = {
        'block_type': 'type_8',
        'length': 5.0,
        'f_params': test8_f2_para
    }
    test8_params = {
        'filename': 'test8.pdf',
        'paper_height': 20.0,
        'paper_width': 2.0,
        'block_params': [test8_block8b_params, test8_block8a_params,
                         test8_block8c_params],
        'transformations': [('scale paper',)]
    }
    Nomographer(test8_params)

    # test 9
    test9_f1_para = {
        'u_min': 0.5,
        'u_max': 1.0,
        'u_min_trafo': 0.5,
        'u_max_trafo': 1.0,
        'f': lambda u: 2 * (u * u - 1.0),
        'g': lambda u: 3 * u * (u + 1.0),
        'h': lambda u: (-u * (u - 1.0)),
        'title': 'p',
        'tick_side': 'left',
        'tick_levels': 4,
        'tick_text_levels': 2
    }
    test9_f2_para = {
        'u_min': 1.0,
        'u_max': 0.75,
        'u_min_trafo': 1.0,
        'u_max_trafo': 0.75,
        'f': lambda v: v,
        'g': lambda v: 1.0,
        'h': lambda v: (-v * v),
        'title': 'h',
        'tick_side': 'right',
        'tick_levels': 3,
        'tick_text_levels': 2
    }
    test9_f3_para = {
        'u_min': 1.0,
        'u_max': 0.5,
        'u_min_trafo': 1.0,
        'u_max_trafo': 0.75,
        'f': lambda w: 2.0 * (2.0 * w + 1.0),
        'g': lambda w: 3.0 * (w + 1.0),
        'h': lambda w: (-(w + 1.0) * (2.0 * w + 1.0)),
        'title': 'L',
        'tick_side': 'left',
        'tick_levels': 4,
        'tick_text_levels': 2
    }
    test9_block_params = {
        'block_type': 'type_9',
        'f1_params': test9_f1_para,
        'f2_params': test9_f2_para,
        'f3_params': test9_f3_para,
        'transform_ini': True,
    }

    test9_params = {
        'filename': 'test9.pdf',
        'paper_height': 10.0,
        'paper_width': 10.0,
        'block_params': [test9_block_params],
        'transformations': [('scale paper',)]
    }
    Nomographer(test9_params)

    # test 9 b
    test9b_f1 = {
        'u_min': 3.0,
        'u_max': 10.0,
        'f': lambda u: 0,
        'g': lambda u: u,
        'h': lambda u: 1.0,
        'title': 'A',
        'title_x_shift': 0.0,
        'title_y_shift': 0.25,
        'scale_type': 'linear',
        'tick_levels': 3,
        'tick_text_levels': 2,
        'tick_side': 'right',
        'tag': 'none',
        'grid': False}

    test9b_f2 = {
        'u_min': 3.0,
        'u_max': 10.0,
        'f': lambda u: 4.0,
        'g': lambda u: u,
        'h': lambda u: 1.0,
        'title': 'B',
        'title_x_shift': 0.0,
        'title_y_shift': 0.25,
        'scale_type': 'linear',
        'tick_levels': 3,
        'tick_text_levels': 2,
        'tick_side': 'right',
        'tag': 'none',
        'grid': False}

    test9b_f3 = {
        'ID': 'none',  # to identify the axis
        'tag': 'none',  # for aligning block wrt others
        'title': 'Grid',
        'title_x_shift': 0.0,
        'title_y_shift': 0.25,
        'title_distance_center': 0.5,
        'title_opposite_tick': True,
        'u_min': 0.0,  # for alignment
        'u_max': 1.0,  # for alignment
        'f_grid': lambda u, v: u + 2.0,
        'g_grid': lambda u, v: 2 * v,
        'h_grid': lambda u, v: 1.0,
        'u_start': 0.0,
        'u_stop': 1.0,
        'v_start': 0.0,
        'v_stop': 1.0,
        'u_values': [0.0, 0.25, 0.5, 0.75, 1.0],
        'v_values': [0.0, 0.25, 0.5, 0.75, 1.0],
        'grid': True
    }
    test9b_block_params = {
        'block_type': 'type_9',
        'f1_params': test9b_f1,
        'f2_params': test9b_f2,
        'f3_params': test9b_f3,
        'transform_ini': False,
    }

    test9b_params = {
        'filename': 'test9b.pdf',
        'paper_height': 10.0,
        'paper_width': 10.0,
        'block_params': [test9b_block_params],
        'transformations': [('rotate', 0.01), ('scale paper',)]
    }
    Nomographer(test9b_params)

    # test 9 c
    test9c_f1 = {
        'u_min': 0.0,
        'u_max': 100.0,
        'f': lambda u: 0,
        'g': lambda u: u,
        'h': lambda u: 1.0,
        'title': 'm',
        'title_x_shift': 0.0,
        'title_y_shift': 0.25,
        'scale_type': 'linear',
        'tick_levels': 3,
        'tick_text_levels': 2,
        'tick_side': 'left',
        'tag': 'none',
        'grid': False}

    test9c_f2 = {
        'u_min': 0.0,
        'u_max': 100.0,
        'f': lambda u: 100.0,
        'g': lambda u: 100.0 - u,
        'h': lambda u: 1.0,
        'title': 'L',
        'title_x_shift': 0.0,
        'title_y_shift': 0.25,
        'scale_type': 'linear',
        'tick_levels': 3,
        'tick_text_levels': 2,
        'tick_side': 'right',
        'tag': 'none',
        'grid': False}

    test9c_f3 = {
        'ID': 'none',  # to identify the axis
        'tag': 'none',  # for aligning block wrt others
        'title': 'Grid',
        'title_x_shift': 0.0,
        'title_y_shift': 0.25,
        'title_distance_center': 0.5,
        'title_opposite_tick': True,
        'u_min': 0.0,  # for alignment
        'u_max': 1.0,  # for alignment
        'f_grid': lambda u, v: 100 * np.cos(u * np.pi / 180.0) / (1.0 + np.cos(u * np.pi / 180.0)),
        'g_grid': lambda u, v: (v * np.sin(u * np.pi / 180.0) + 100.0 * np.cos(u * np.pi / 180.0)) / (
            1.0 + np.cos(u * np.pi / 180.0)),
        'h_grid': lambda u, v: 1.0,
        'u_start': 15.0,
        'u_stop': 75.0,
        'v_start': 0.0,
        'v_stop': 100.0,
        'u_values': [15.0, 30.0, 45.0, 60.0, 75.0],
        'v_values': [0.0, 20.0, 40.0, 60.0, 80.0, 100.0],
        'grid': True,
        'text_prefix_u': r'$\theta$=',
        'text_prefix_v': r'k=',
        'text_distance': 0.5,
    }
    test9c_block_params = {
        'block_type': 'type_9',
        'f1_params': test9c_f1,
        'f2_params': test9c_f2,
        'f3_params': test9c_f3,
        'transform_ini': False,
    }

    test9c_params = {
        'filename': 'test9c.pdf',
        'paper_height': 10.0,
        'paper_width': 10.0,
        'block_params': [test9c_block_params],
        'transformations': [('rotate', 0.01), ('polygon',), ('scale paper',)]
    }
    Nomographer(test9c_params)

    # test 9 d
    test9d_f1 = {
        'u_min': 5.0,
        'u_max': 15.0,
        'f': lambda u: 0,
        'g': lambda s: s,
        'h': lambda u: 1.0,
        'title': 's',
        'title_x_shift': 0.0,
        'title_y_shift': 0.25,
        'scale_type': 'linear',
        'tick_levels': 3,
        'tick_text_levels': 2,
        'tick_side': 'left',
        'tag': 'none',
        'grid': False}

    test9d_f2 = {
        'u_min': 0.0,
        'u_max': 10.0,
        'f': lambda u: 1.0,
        'g': lambda u: -u,
        'h': lambda u: 1.0,
        'title': 'v',
        'title_x_shift': 0.0,
        'title_y_shift': 0.25,
        'scale_type': 'linear',
        'tick_levels': 3,
        'tick_text_levels': 2,
        'tick_side': 'right',
        'tag': 'none',
        'grid': False}

    test9d_f3 = {
        'ID': 'none',  # to identify the axis
        'tag': 'none',  # for aligning block wrt others
        'title': 'Grid',
        'title_x_shift': 0.0,
        'title_y_shift': 0.25,
        'title_distance_center': 0.5,
        'title_opposite_tick': True,
        'u_min': 0.0,  # for alignment
        'u_max': 1.0,  # for alignment
        'f_grid': lambda u, v: u / (u + 1.0),
        'g_grid': lambda u, v: 0.5 * v * u ** 2 / (u + 1.0),
        'h_grid': lambda u, v: 1.0,
        'u_start': 1.0,
        'u_stop': 3.0,
        'v_start': 0.0,
        'v_stop': 4.0,
        'u_values': [1.0, 2.0, 3.0],
        'v_values': [0.0, 1.0, 2.0, 3.0, 4.0],
        'text_prefix_u': r't=',
        'text_prefix_v': r'a=',
        'text_distance': 0.45,
        'grid': True,
        'v_texts_u_start': True,
        'v_texts_u_stop': False,
    }
    test9d_block_params = {
        'block_type': 'type_9',
        'f1_params': test9d_f1,
        'f2_params': test9d_f2,
        'f3_params': test9d_f3,
        'transform_ini': False,
    }

    test9d_params = {
        'filename': 'test9d.pdf',
        'paper_height': 10.0,
        'paper_width': 10.0,
        'block_params': [test9d_block_params],
        'transformations': [('rotate', 0.01), ('polygon',), ('scale paper',)]
    }
    Nomographer(test9d_params)

    # test type 10
    test_10_f1 = {
        'u_min': -25.0,
        'u_max': 0.0,
        'function': lambda u: u,
        'title': 'F1',
        'tag': 'none',
        'tick_side': 'right',
        'tick_levels': 2,
        'tick_text_levels': 1
    }
    test_10_f2 = {
        'u_min': -5.0,
        'u_max': 0.0,
        'function': lambda u: u,
        'title': 'F2',
        'tag': 'none',
        'tick_side': 'right',
        'tick_levels': 2,
        'tick_text_levels': 1
    }
    test_10_f3 = {
        'u_min': 0.5,
        'u_max': 5.0,
        'function_3': lambda u: u,
        'function_4': lambda u: u ** 2,
        'title': 'F3',
        'tag': 'none',
        'tick_side': 'right',
        'tick_levels': 2,
        'tick_text_levels': 1
    }
    test10_block_params = {
        'block_type': 'type_10',
        'f1_params': test_10_f1,
        'f2_params': test_10_f2,
        'f3_params': test_10_f3,
    }

    test10_params = {
        'filename': 'test10.pdf',
        'paper_height': 10.0,
        'paper_width': 10.0,
        'block_params': [test10_block_params],
        'transformations': [('rotate', 0.01), ('scale paper',)]
    }
    Nomographer(test10_params)



# -*- coding: utf-8 -*-
#
#    This file is part of PyNomo -
#    a program to create nomographs with Python (https://github.com/lefakkomies/pynomo)
#
#    Copyright (C) 2007-2019  Leif Roschier  <lefakkomies@users.sourceforge.net>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from scipy import optimize
import numpy as np
import pyx
import math
import time


class Nomo_Grid_Box(object):
    """
    class to calculate "grid boxes" according to (East)-German nomographic tradition
    """

    def __init__(self, params={}):
        """
        params = definitions
        """
        params_default_values = {'width': 10.0,
                                 'height': 10.0,
                                 'mirror_x': False,
                                 'mirror_y': False,
                                 'u_func': lambda u: u,
                                 'v_func': lambda x, para: x + para,
                                 'line_func': lambda x: x,  # function for centerline assuming square
                                 'draw_line': True,
                                 'w_func': lambda x: x,  # function for w axis
                                 'w_func_inv': lambda x: x,  # inverse function for w axis
                                 'w_min': 0.1,
                                 'w_max': 1.0,
                                 'manual_w_scale': False,
                                 'wd_func': lambda x: x,  # function for w axis
                                 'wd_func_inv': lambda x: x,  # inverse function for w axis
                                 'wd_min': 0.1,
                                 'wd_max': 1.0,
                                 'manual_wd_scale': False,
                                 'v_min': 0.1,  # not used typically
                                 'v_max': 1.0,  # not used typically
                                 'x_min': 0.1,
                                 'x_max': 1.0,
                                 'manual_x_scale': False,  # if inital x_scale is set manually, use two values above
                                 'u_values': [0.0, 0.25 * math.pi / 2, 0.5 * math.pi / 2, 0.75 * math.pi / 2,
                                              1.0 * math.pi / 2],
                                 'v_values': [0.0, 0.25 * math.pi / 2, 0.5 * math.pi / 2, 0.75 * math.pi / 2,
                                              1.0 * math.pi / 2],
                                 'wd_values': [],
                                 'w_values': [],
                                 'u_tag': 'none',  # manual labels
                                 'w_tag': 'none',
                                 'wd_tag': 'none',
                                 'u_reference': False,  # manual labels
                                 'v_reference': False,
                                 'w_reference': False,
                                 'wd_reference': False,
                                 'scale_type_u': 'manual line',
                                 'scale_type_w': 'linear',
                                 'scale_type_wd': 'linear',
                                 'scale_type_v': 'manual line',  # change in definition by own risk
                                 'u_title': '',
                                 'v_title': '',
                                 'w_title': '',
                                 'wd_title': '',
                                 'u_tick_side': None,  # can be overriden
                                 'u_tick_levels': 0,
                                 'u_tick_text_levels': 0,
                                 'v_tick_side': 'right',
                                 'v_tick_levels': 0,
                                 'v_tick_text_levels': 0,
                                 'w_tick_side': 'right',
                                 'w_tick_levels': 0,
                                 'w_tick_text_levels': 0,
                                 'wd_tick_side': 'left',
                                 'wd_tick_levels': 0,
                                 'wd_tick_text_levels': 0,
                                 'horizontal_guide_nr': 51,
                                 'horizontal_guides': True,
                                 'vertical_guide_nr': 51,
                                 'vertical_guides': True,
                                 'u_title_opposite_tick': False,
                                 'v_title_opposite_tick': False,
                                 'wd_title_opposite_tick': False,
                                 'w_title_opposite_tick': False,
                                 'u_title_distance_center': 1.5,
                                 'v_title_distance_center': 1.5,
                                 'wd_title_distance_center': 1.5,
                                 'w_title_distance_center': 1.5,
                                 'u_title_draw_center': True,
                                 'v_title_draw_center': True,
                                 'wd_title_draw_center': True,
                                 'w_title_draw_center': True,
                                 'u_text_format': r"$%4.4g$ ",
                                 'v_text_format': r"$%4.4g$ ",
                                 'u_align_func': lambda u: u,
                                 'wd_align_func': lambda u: u,
                                 'w_align_func': lambda u: u,
                                 'u_scale_opposite': False,
                                 'v_manual_axis_data': None,
                                 'u_manual_axis_data': None,
                                 'v_text_distance': 0.25,
                                 'u_align_x_offset': 0.0,
                                 'wd_align_x_offset': 0.0,
                                 'u_align_y_offset': 0.0,
                                 'wd_align_y_offset': 0.0,
                                 'u_axis_color': pyx.color.rgb.black,
                                 'u_text_color': pyx.color.rgb.black,
                                 'u_title_color': pyx.color.rgb.black,
                                 'v_axis_color': pyx.color.rgb.black,
                                 'v_text_color': pyx.color.rgb.black,
                                 'v_title_color': pyx.color.rgb.black,
                                 'wd_axis_color': pyx.color.rgb.black,
                                 'wd_text_color': pyx.color.rgb.black,
                                 'wd_title_color': pyx.color.rgb.black,
                                 'allow_additional_v_scale': False,  # to draw additional scale as atom
                                 # isopleths do not work
                                 'v_scale_u_value': 1.0,  # this value sets additional v_scale
                                 'x_func': lambda u, v: u + v  # gives x as a function of u and v
                                 }
        self.params = params_default_values
        self.params.update(params)
        # initial guesses
        self.u_func = self.params['u_func']
        self.v_func = self.params['v_func']
        self._build_v_lines_(self.v_func)
        self._calc_bound_box_ini_()
        self._build_u_lines_(self.u_func)
        # debug by looking pdf
        # self._draw_debug_ini_()
        # build scaled versions
        self._scale_and_mirror_()
        # self._draw_debug_ini_('after.pdf')

        self.give_u_data()
        self.give_v_data()
        self.give_w_data()
        self.give_wd_data()

    def give_u_data(self):
        """
        gives F(u),G(u)-functions and other params (left coordinate)
        """
        # calc func
        u_func = self.u_func  # defined in scaling
        u_manual_axis_data = {}
        if self.params['u_manual_axis_data'] == None:
            for u_value in self.params['u_values']:
                u_manual_axis_data[u_value] = self.params['u_text_format'] % u_value
        else:
            u_manual_axis_data = self.params['u_manual_axis_data']
        if self.params['u_scale_opposite']:
            x_coordinate = self.x_right
        else:
            x_coordinate = self.x_left
        self.params_u = {
            'u_min': min(self.params['u_values']),
            'u_max': max(self.params['u_values']),
            'F': lambda u: x_coordinate,  # x-coordinate
            'G': u_func,  # y-coordinate
            'title': self.params['u_title'],
            # 'linear' 'log' 'manual point' 'manual line'
            'scale_type': self.params['scale_type_u'],
            'manual_axis_data': u_manual_axis_data,
            'tick_side': self.params['u_tick_side'],
            'tag': self.params['u_tag'],  # for aligning block wrt others
            'reference': self.params['u_reference'],
            # not really used, yet
            'tick_levels': self.params['u_tick_levels'],
            # not really used, yet
            'tick_text_levels': self.params['u_tick_text_levels'],
            'title_opposite_tick': self.params['u_title_opposite_tick'],
            'title_distance_center': self.params['u_title_distance_center'],
            'title_draw_center': self.params['u_title_draw_center'],
            'align_func': self.params['u_align_func'],
            'align_x_offset': self.params['u_align_x_offset'],
            'align_y_offset': self.params['u_align_y_offset'],
            # 'text_format':self.params['u_text_format'],
            'axis_color': self.params['u_axis_color'],
            'text_color': self.params['u_text_color'],
            'title_color': self.params['u_title_color'],
        }
        return self.params_u

    def give_v_data(self):
        """
        givesgives F(u),G(u)-functions and other params (top coordinate)
        function is linear and values are set manually to correct places
        """
        v_manual_axis_data = {}
        # let's go through all lines and find x-coordinate of top points.
        for idx1, v_line in enumerate(self.v_lines):
            x_max, y_max = v_line[0]
            for idx2, (x, y) in enumerate(v_line):
                if y > y_max:
                    x_max = x
                    y_max = y
            v_value = self.params['v_values'][idx1]
            # v_manual_axis_data[v_value]='%f'%x_max
            v_manual_axis_data[x_max] = '%3.2f' % v_value
        # in case one wants better v-scale
        if self.params['allow_additional_v_scale']:
            v_min = self.params['v_min']
            v_max = self.params['v_max']
            # this value has to be set manually
            u_value = self.params['v_scale_u_value']
            def f_v(v): return self.x_func(u_value, v)
            def g_v(u): return self.u_func(u_value)
        else:  # assuming manual data
            v_min = self.x_left
            v_max = self.x_right
            def f_v(u): return u
            def g_v(u): return self.x_top
        self.params_v = {
            #            'u_min':self.x_left,
            #            'u_max':self.x_right,
            #            'F':lambda u:u, # x-coordinate
            #            'G':lambda u:self.y_top, # y-coordinate
            'u_min': v_min,
            'u_max': v_max,
            'F': f_v,
            'G': g_v,
            'title': self.params['v_title'],
            # 'scale_type':'manual line', #this have to be hard coded
            'scale_type': self.params['scale_type_v'],
            'manual_axis_data': v_manual_axis_data,
            'tick_side': self.params['v_tick_side'],
            'tag': 'none',  # this axis should not be aligned
            'reference': self.params['v_reference'],
            # not really used, yet
            'tick_levels': self.params['v_tick_levels'],
            # not really used, yet
            'tick_text_levels': self.params['v_tick_text_levels'],
            'title_opposite_tick': self.params['v_title_opposite_tick'],
            'title_distance_center': self.params['v_title_distance_center'],
            'title_draw_center': self.params['v_title_draw_center'],
            'text_format': self.params['v_text_format'],
            'axis_color': self.params['v_axis_color'],
            'text_color': self.params['v_text_color'],
            'title_color': self.params['v_title_color'],
        }
        return self.params_v

    def give_w_data(self):
        """
        gives x(w),y(w)-functions and w_min and w_max values.
        w = f2(f3(x))
        x = bottom coordinate of initial axes
        f3 = line func
        f2 = additional transformation
        """
        f3 = self.params['line_func']
        f2 = self.params['w_func']
        f2_inv = self.params['w_func_inv']
        w_manual_axis_data = {}
        if self.params['scale_type_w'] == 'manual line':
            for w_value in self.params['w_values']:
                w_manual_axis_data[w_value] = '%f' % w_value

        w_min = f2_inv(f3(self.x_left_ini))
        w_max = f2_inv(f3(self.x_right_ini))
        w_diff = w_max - w_min
        if self.params['mirror_y'] == True:
            y_factor = -1.0
        else:
            y_factor = 1.0
        self.params_w = {
            'u_min': w_min,  # this is w_min
            'u_max': w_max,  # this is w_max
            'F': lambda w: self.x_right,  # x-coordinate
            'G': lambda w: self.y_bottom + (f2(w) - f2(w_min)) / (f2(w_max) - f2(w_min)) \
            * self.params['height'] * y_factor,  # y-coordinate
            'title': self.params['w_title'],
            'scale_type': self.params['scale_type_w'],
            'manual_axis_data': w_manual_axis_data,
            'tick_side': self.params['w_tick_side'],
            'tag': self.params['w_tag'],  # for aligning block wrt others
            'reference': self.params['w_reference'],
            'tick_levels': self.params['w_tick_levels'],
            'tick_text_levels': self.params['w_tick_text_levels'],
            'title_opposite_tick': self.params['w_title_opposite_tick'],
            'title_distance_center': self.params['w_title_distance_center'],
            'title_draw_center': self.params['w_title_draw_center'],
            'align_func': self.params['w_align_func'],
        }
        return self.params_w

    def give_wd_data(self):
        """
        gives x(wd),y(wd)-functions and wd_min and wd_max values. (bottom coordinate)
        """
        wd_manual_axis_data = {}
        if self.params['scale_type_wd'] == 'manual line':
            for wd_value in self.params['wd_values']:
                wd_manual_axis_data[wd_value] = '%f' % wd_value
        f1 = self.params['wd_func']
        f1_inv = self.params['wd_func_inv']
        wd_min = f1_inv(self.x_left_ini)
        wd_max = f1_inv(self.x_right_ini)
        self.params_wd = {
            'u_min': wd_min,
            'u_max': wd_max,
            'F': lambda u: f1(u) * self.x_factor,  # x-coordinate
            'G': lambda u: self.y_bottom,  # y-coordinate
            'title': self.params['wd_title'],
            'scale_type': self.params['scale_type_wd'],
            'manual_axis_data': wd_manual_axis_data,
            'tick_side': self.params['wd_tick_side'],
            'tag': self.params['wd_tag'],  # for aligning block wrt others
            'reference': self.params['wd_reference'],
            'tick_levels': self.params['wd_tick_levels'],
            'tick_text_levels': self.params['wd_tick_text_levels'],
            'title_opposite_tick': self.params['wd_title_opposite_tick'],
            'title_distance_center': self.params['wd_title_distance_center'],
            'title_draw_center': self.params['wd_title_draw_center'],
            'align_func': self.params['wd_align_func'],
            'align_x_offset': self.params['wd_align_x_offset'],
            'align_y_offset': self.params['wd_align_y_offset'],
            'axis_color': self.params['wd_axis_color'],
            'text_color': self.params['wd_text_color'],
            'title_color': self.params['wd_title_color'],
        }
        return self.params_wd

    def give_contours(self):
        """
        gives contours to be drawn
        """
        pass

    def _scale_and_mirror_(self):
        """
        scales everything to width and height
        """
        if self.params['mirror_x'] == True:
            mirror_x = -1.0
        else:
            mirror_x = 1.0
        if self.params['mirror_y'] == True:
            mirror_y = -1.0
        else:
            mirror_y = 1.0
        # scales lines
        y_factor = mirror_y * self.params['height'] / self.BB_height_ini
        x_factor = mirror_x * self.params['width'] / self.BB_width_ini
        for idx1, u_line in enumerate(self.u_lines):
            for idx2, (x, y) in enumerate(u_line):
                x_new = x * x_factor
                y_new = y * y_factor
                self.u_lines[idx1][idx2] = (x_new, y_new)
        for idx1, v_line in enumerate(self.v_lines):
            for idx2, (x, y) in enumerate(v_line):
                x_new = x * x_factor
                y_new = y * y_factor
                self.v_lines[idx1][idx2] = (x_new, y_new)
        # scale functions
        self.u_func = lambda u: self.params['u_func'](u) * y_factor
        self.v_func = lambda x, v: self.params['v_func'](
            x / x_factor, v) * y_factor
        self.x_func = lambda u, v: self.params['x_func'](u, v) * x_factor
        self.x_left = self.x_left_ini * x_factor
        self.x_right = self.x_right_ini * x_factor
        self.y_bottom = self.y_bottom_ini * y_factor
        self.y_top = self.y_top_ini * y_factor
        self.x_factor = x_factor
        self.y_factor = y_factor

    def _build_u_scale_(self):
        """
        vertical scale on the left
        """
        pass

    def _build_v_scale_(self):
        """
        horizontal scale on the top
        """
        pass

    def _build_u_lines_(self, u_func):
        """
        lines starting from u scale
        """
        self.u_lines = []
        self.u_sections = []
        # g=self.u_func
        g = u_func
        for u in self.params['u_values']:
            line = [(self.x_left_ini, g(u))]
            line.append((self.x_left_ini, g(u)))
            line.append((self.x_right_ini, g(u)))
            section = [(self.x_left_ini, g(u), self.x_right_ini, g(u))]
            self.u_lines.append(line)
            self.u_sections.append(section)

    def _build_v_lines_(self, v_func):
        """
        build lines starting from top scale
        """
        self.v_lines = []
        self.v_sections = []
        for v in self.params['v_values']:
            line, sections = self._build_v_line_(v_func=v_func, v=v)
            self.v_lines.append(line)
            self.v_sections.append(sections)

    def _build_v_line_(self, v_func, v=1.0):
        """
        line starting from x scale
        code copied originally form nomo_axis_func.py: _calculate_points_
        v_func is the function(x,p)
        p is the parametric value of the top scale
        """
        # find top and bottom lines
        max_fu = self.u_func(self.params['u_values'][0])
        min_fu = max_fu
        for u in self.params['u_values']:
            fu = self.u_func(u)
            if fu > max_fu:
                max_fu = fu
            if fu < min_fu:
                min_fu = fu
        # functions to find x-values in paper
        func2 = v_func

        def func_top(x):
            value = func2(x.astype(complex), v)
            if value.imag > 0:
                return 1e10  # big number
            if value.imag < 0:
                return -1e10  # big number
            return (func2(x, v) - max_fu) ** 2

        def func_bottom(x):
            value = func2(x.astype(complex), v)
            if value.imag > 0:
                return 1e10  # big number
            else:
                return (func2(x, v) - min_fu) ** 2

        # func_top=lambda x:((func2(x.astype(complex),v)-max_fu)**2).real+1e8*((func2(x.astype(complex),v)-max_fu)**2).imag # minimum at height
        # func_bottom=lambda x:((func2(x.astype(complex),v)-min_fu)**2).real+1e8*((func2(x.astype(complex),v)-min_fu)**2).imag # minimum at 0.0
        def f(x): return x
        def g(x): return func2(x, v)
        # find point of scale to meet point 1.0
        x_guess_top = 1.0
        x_guess_bottom = 1.0
        if self.params['manual_x_scale'] == True:
            mean_x = (self.params['x_min'] + self.params['x_max']) / 2.0
            x_guess_top = mean_x
            x_guess_bottom = mean_x
        x_top = optimize.fmin(
            func_top, [x_guess_top], disp=0, ftol=1e-5, xtol=1e-5)[0]
        x_bottom = optimize.fmin(
            func_bottom, [x_guess_bottom], disp=0, ftol=1e-5, xtol=1e-5)[0]
        # print "x_top %f"%x_top
        # print "x_bottom %f" % x_bottom
        # print "g(x_top) %f"%g(x_top)
        # print "g(x_bottom) %f" %g(x_bottom)
        if self.params['manual_x_scale'] == True:
            x_min_limit = self.params['x_min']
            x_max_limit = self.params['x_max']
            if x_min_limit > x_max_limit:
                x_min_limit, x_max_limit = x_max_limit, x_min_limit
            if x_top > x_max_limit:
                x_top = x_max_limit
            if x_top < x_min_limit:
                x_top = x_min_limit
            if x_bottom > x_max_limit:
                x_bottom = x_max_limit
            if x_bottom < x_min_limit:
                x_bottom = x_min_limit
        start = min(x_top, x_bottom)
        stop = max(x_top, x_bottom)
        # if self.params['manual_x_scale']==True:
        #    start=min(self.params['x_min'],self.params['x_max'])
        #    stop=max(self.params['x_min'],self.params['x_max'])
        du = np.fabs(stop - start) * 1e-12
        # approximate line length is found
        line_length_straigth = max_fu - min_fu
        sections = 200.0  # number of sections
        section_length = line_length_straigth / sections
        line = [(f(start), g(start))]
        line.append((f(start), g(start)))
        u = start
        count = 1
        # print "v:%g"%v
        while True:
            if u < stop:
                count = 1
                dx = (f(u + du) - f(u))
                dy = (g(u + du) - g(u))
                dl = np.sqrt(dx ** 2 + dy ** 2)
                if dl > 0:
                    delta_u = du * section_length / dl
                else:
                    delta_u = du
                # let's calculate actual length
                # and iterate until length is in factor 2 from target
                while True:
                    count = count + 1
                    delta_x = f(u + delta_u) - f(u)
                    delta_y = g(u + delta_u) - g(u)
                    delta_l = np.sqrt(delta_x ** 2 + delta_y ** 2)
                    if delta_l > 2.0 * section_length:
                        delta_u = delta_u * 0.999
                        # print "v:%g, delta_x:%g delta_y:%g delta_l:%g, section_length:%g, delta_u math.pienenee:%g"%(v,delta_x,delta_y,delta_l,section_length,delta_u)
                    else:
                        if delta_l < section_length / 2.0:
                            delta_u = delta_u * 1.001
                            # print "v:%g, delta_x:%g delta_y:%g delta_l:%g, section_length:%g, delta_u kasvaa:%g"%(v,delta_x,delta_y,delta_l,section_length,delta_u)
                    if delta_l <= 2 * section_length and delta_l >= 0.5 * section_length:
                        # print "selvitty"
                        break
                    if count > 1e3:  # cancel if not solution
                        print("Warning no solution found in contour")
                        break

                u += delta_u
                # print u,stop
                if (u < stop and count < 1e3):
                    line.append((f(u), g(u)))
                else:
                    u = stop  # stop looping of contour
                    # print "count %f"%count
            else:
                line.append((f(stop), g(stop)))
                # print count
                # print len(line)
                break
        # calculate sections
        sections = []
        for index, (x, y) in enumerate(line):
            if index > 1:
                sections.append((x, y, prev_x, prev_y))
            prev_x = x
            prev_y = y
        return line, sections

    def _calc_bound_box_ini_(self):
        """
        calculates bounding initial bounding box
        """
        (x_0, y_0) = self.v_lines[0][0]
        x_left = x_0
        y_top = max(self.u_func(np.array(self.params['u_values'])))
        x_right = x_0
        y_bottom = min(self.u_func(np.array(self.params['u_values'])))
        for line in self.v_lines:
            for x, y in line:
                if x < x_left:
                    x_left = x
                if x > x_right:
                    x_right = x
                if y < y_bottom:
                    y_bottom = y
                if y > y_top:
                    y_top = y
                    # print "y_top %f"%y_top
        # print x_left,x_right,y_bottom,y_top
        if self.params['manual_x_scale'] == True:
            self.x_left_ini = self.params['x_min']
            self.x_right_ini = self.params['x_max']
        else:
            self.x_left_ini = x_left
            self.x_right_ini = x_right
        self.y_top_ini = y_top
        self.y_bottom_ini = y_bottom
        # self.BB_width_ini=x_right-x_left
        self.BB_width_ini = abs(self.x_left_ini - self.x_right_ini)
        self.BB_height_ini = y_top - y_bottom
        return x_left, x_right, y_bottom, y_top

    #    def _build_box_(self):
    #        """
    #        box around structure
    #        """
    #        pass
    #
    #    def _build_diagonal_line(self):
    #        """
    #        diagonal line to make 90 degree angle
    #        """
    #        pass

    def _draw_debug_ini_(self, filename='nomo_grid_test_debug.pdf'):
        """
        draws lines for debugging purposes, initial figure
        """
        cc = pyx.canvas.canvas()
        x00, y00 = self.u_lines[0][0]
        line = pyx.path.path(pyx.path.moveto(x00, y00))
        for u_line in self.u_lines:
            x0, y0 = u_line[0]
            line.append(pyx.path.moveto(x0, y0))
            for x, y in u_line:
                line.append(pyx.path.lineto(x, y))
        for v_line in self.v_lines:
            x0, y0 = v_line[0]
            line.append(pyx.path.moveto(x0, y0))
            for x, y in v_line:
                line.append(pyx.path.lineto(x, y))

        cc.stroke(line, [pyx.style.linewidth.normal])
        cc.writePDFfile(filename)


if __name__ == '__main__':
    def f1(x, u):
        return np.log(x / (x - u / 100.0)) / np.log(1 + u / 100.0)

    params = {'width': 10.0,
              'height': 10.0,
              'u_func': lambda u: u,
              'v_func': f1,
              'u_values': [10.0, 15.0, 20.0, 25.0, 30.0, 40.0, 50.0, 60.0],
              'v_values': [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0],
              'v_tick_side': 'left',
              }
    tic = time.time()
    test = Nomo_Grid_Box(params=params)
    toc = time.time()
    # print toc - tic, ' has elapsed'

    manual_axis_data = {1.0: 'first',
                        2.0: 'second',
                        3.0: 'third',
                        3.1415: r'$\pi$',
                        4.0: 'fourth',
                        5.0: 'fifth',
                        6.0: 'sixth',
                        7.0: 'seventh',
                        8.0: 'eigth',
                        9.0: 'nineth',
                        10.0: 'tenth'}
